using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;

namespace Jade.DataMining.Classifier
{
    /// <summary>
    /// 
    /// </summary>
    public class ChiSquareFilter : ModelTraining
    {

        public string Method { get; private set; }

        public ChiSquareFilter(bool loadFromFile = false)
        {
            this.Method = ClassiferSetting.FilterSetting.Method;
            this.IsTrain = loadFromFile;
            if (loadFromFile)
            {
                this.Load();
            }
        }

        public List<List<double>> ChiTable { get; set; }

        public List<List<double>> ATable { get; set; }

        public List<int> IdMap { get; set; }

        /// <summary>
        /// ѵ
        ///  㹫ʽ:
        ///  t: term
        ///  c: category
        ///  X^2(t, c) =   N * (AD - CB)^2
        ///             ____________________
        ///             (A+C)(B+D)(A+B)(C+D)
        ///  A,B,C,D is doc-count
        ///  A:     belong to c,     include t
        ///  B: Not belong to c,     include t
        ///  C:     belong to c, Not include t
        ///  D: Not belong to c, Not include t
        /// 
        ///  B = t's doc-count - A
        ///  C = c's doc-count - A
        ///  D = N - A - B - C
        /// and score of t can be calculated by next 2 formulations:
        /// X^2(t) = sigma p(ci)X^2(t,ci) (avg)
        ///            i
        /// X^2(t) = max { X^2(t,c) }     (max)
        /// """
        /// </summary>
        /// <param name="matrix"></param>
        public void TrainFilter(Matrix matrix)
        {
            if (matrix.RowsCount != matrix.Categories.Count)
            {
                throw new Exception("ERROR!,matrix.RowsCount shoud be equal to matrix.Categories.Count");
            }

            var distinctCategories = matrix.Categories.Distinct().ToList();
            distinctCategories.Sort();

            //#create a table stores X^2(t, c)
            // #create a table stores A(belong to c, and include t ά
            ChiTable = new List<List<double>>();
            var data = new List<double>();
            for (var j = 0; j < matrix.ColsCount; j++)
            {
                data.Add(0);
            }

            for (var i = 0; i < distinctCategories.Count; i++)
            {
                ChiTable.Add(data.AsReadOnly().ToList());
            }

            // atable [category][term] - count
            ATable = ChiTable.AsReadOnly().ToList();

            for (var row = 0; row < matrix.RowsCount; row++)
            {
                for (var col = matrix.Rows[row]; col < matrix.Rows[row + 1]; col++)
                {
                    var categoryId = matrix.Categories[row];
                    var termId = matrix.Cols[col];
                    ATable[categoryId][termId] += 1;
                }
            }

            // ĵ
            var n = matrix.RowsCount;

            // 㿨
            for (var t = 0; t < matrix.ColsCount; t++)
            {
                for (var cc = 0; cc < distinctCategories.Count; cc++)
                {
                    var a = ATable[distinctCategories[cc]][matrix.Cols[t]]; // ڷccҰtĿ
                    var b = GlobalInfo.Instance.IdToDocCount[t] - a; // tǲڷĵ = t-ccĿ
                    var c = GlobalInfo.Instance.ClassToDocCount[distinctCategories[cc]] - a;  // ڷcctĿ = cĿ - ڷt
                    var d = n - a - b - c; // ȲcֲtĿ
                    //#get X^2(t, c)
                    var numberator = (n) * (a * d - c * b) * (a * d - c * b) + 1;
                    var denominator = (a + c) * (b + d) * (a + b) * (c + d) + 1;
                    ChiTable[distinctCategories[cc]][t] = numberator / denominator;
                }
            }

            // chiScore[t][2]  : chiScore[t][0] = score,chiScore[t][1]  = colIndex
            var chiScore = new List<List<double>>();
            for (var i = 0; i < matrix.ColsCount; i++)
            {
                var c = new List<double>();
                for (var j = 0; j < 2; j++)
                {
                    c.Add(0);
                }
                chiScore.Add(c);
            }

            // avg ʱ յ÷ X^2(t) = sigma p(ci)X^2(t,ci)  p(ci)Ϊ
            if (this.Method == "avg")
            {
                // pc [category] - categoyCount/n
                var priorC = new double[distinctCategories.Count + 1];
                for (var i = 0; i < distinctCategories.Count; i++)
                {
                    priorC[distinctCategories[i]] = (double)GlobalInfo.Instance.ClassToDocCount[distinctCategories[i]] / n;
                }

                // ÷
                for (var t = 0; t < matrix.ColsCount; t++)
                {
                    chiScore[t][1] = t;
                    for (var c = 0; c < distinctCategories.Count; c++)
                    {
                        chiScore[t][0] += priorC[distinctCategories[c]] * ChiTable[distinctCategories[c]][t];
                    }
                }
            }
            else
            {
                //method == "max"
                // calculate score of each t
                for (var t = 0; t < matrix.ColsCount; t++)
                {
                    chiScore[t][1] = t;
                    // ȡֵ
                    for (var c = 0; c < distinctCategories.Count; c++)
                    {
                        if (chiScore[t][0] < ChiTable[distinctCategories[c]][t])
                            chiScore[t][0] = ChiTable[distinctCategories[c]][t];
                    }
                }

            }

            // Ƚϵ÷
            chiScore.Sort(new ScoreCompare());
            chiScore.Reverse();

            #region
            var idMap = new int[matrix.ColsCount];

            // add un-selected feature-id to idmap
            for (var i = (int)(ClassiferSetting.FilterSetting.Rate * chiScore.Count); i < chiScore.Count; i++)
            {
                // δѡеıΪ-1
                var termId = chiScore[i][1];
                idMap[(int)termId] = -1;
            }
            var offset = 0;
            for (var t = 0; t < matrix.ColsCount; t++)
            {
                if (idMap[t] < 0)
                {
                    offset += 1;
                }
                else
                {
                    idMap[t] = t - offset;
                    GlobalInfo.Instance.NewIdToId[t - offset] = t;
                }
            }

            this.IdMap = new List<int>(idMap);
            #endregion

            StringBuilder stringBuilder = new StringBuilder();
            stringBuilder.AppendLine("chiSquare info:");
            stringBuilder.AppendLine("=======selected========");
            for (var i = 0; i < chiScore.Count; i++)
            {
                if (i == (int)(ClassiferSetting.FilterSetting.Rate * chiScore.Count))
                {
                    stringBuilder.AppendLine("========unselected=======");
                }
                var term = GlobalInfo.Instance.IdToTerm[(int)chiScore[i][1]];
                var score = chiScore[i][0];
                stringBuilder.AppendLine(string.Format("{0} {1}", term, score));
            }
            File.WriteAllText(ClassiferSetting.FilterSetting.LogPath, stringBuilder.ToString());

            GlobalInfo.Instance.Save();

            this.Save();

            this.IsTrain = true;
        }

        public void SampleFilter(Matrix matrix)
        {
            if (!this.IsTrain)
            {
                throw new Exception("ѡѵģ");
            }
            //#filter sample
            var newCols = new List<int>();
            var newVals = new List<int>();
            for (var c = 0; c < matrix.Cols.Count; c++)
            {
                if (IdMap[matrix.Cols[c]] >= 0)
                {
                    newCols.Add(matrix.Cols[c]);
                    newVals.Add(matrix.Vals[c]);
                }
            }
            matrix.Vals = newVals;
            matrix.Cols = newCols;
        }

        #region Overrides of ModelTraining

        public override void Load()
        {
            try
            {
                Console.WriteLine("ؿģ͡");
                this.IdMap = (List<int>)SerializeHelper.Load(typeof(List<int>), ClassiferSetting.FilterSetting.ModelPath);
            }
            catch
            {
                Console.WriteLine("ؿģʧ");
            }
        }

        public override void Save()
        {
            try
            {
                SerializeHelper.Save(this.IdMap, ClassiferSetting.FilterSetting.ModelPath);
            }
            catch
            {
                Console.WriteLine("濨ģʧ");
            }
        }

        #endregion
    }
}